import re 
from . import constant
from .code_ import Code
import copy
from .root import Root
from .main import Main
from .bioflowinsighterror import BioFlowInsightError
from .outils import remove_jumps_inbetween_parentheses, replace_group1, get_dico_from_tab_from_id, check_if_element_in_tab_rocrate




class Subworkflow(Main):
    def __init__(self, code, nextflow_file, name):
        Main.__init__(self, code, nextflow_file)
        self.name = name.replace("'", "").replace('"', '')
        self.alias = self.name
        self.printed_name = self.name
        #These are the different parts of of a subworkflow -> work corresponds to the main 
        self.take = []
        self.takes_channels = []
        self.work = None
        self.emit = []
        self.call = []

        #These are probably to remove
        self.initialised = False
        self.later_emits = []
        self.number_times_called = 0

        self.called_by = []#List of calls
        self.number_times_copied = 0

    def copy(self):
        sub = copy.copy(self)
        sub.alias = self.name
        sub.printed_name = self.printed_name 
        sub.take = []
        sub.takes_channels = []
        sub.work = None
        sub.emit = []
        sub.call = []
        sub.initialised = False
        sub.later_emits = []
        sub.number_times_called = 0
        sub.called_by = []
        num = self.number_times_copied
        self.number_times_copied+=1
        return sub, num
    
    def add_to_calls(self, call):
        self.called_by.append(call)

    def get_call(self):
        if(len(self.called_by)!=1):
            raise Exception("This shouldn't happen")
        return self.called_by[0]
    
    #TODO make sure this is uptodate
    def get_calls_by_name(self, name):            
        tab = []
        for call in self.root.get_calls_same_level():
            #call.initialise()
            for c in call.get_all_calls():
                if(c.first_element_called.get_alias()==name):
                    tab.append(c)
        #Here it is important that BioFlow-Insight is not a Nextflow verificator
        #Here i'm checking the call inside the block
        if(len(tab)==0):
            for call in self.root.get_calls_inside_level():
                #call.initialise()
                for c in call.get_all_calls():
                    if(c.first_element_called.get_alias()==name):
                        tab.append(c)
        return tab
        

    def add_to_emits(self, emit):
        self.later_emits.append(emit)
    
    def get_later_emits(self):
        return self.later_emits

    def get_code_with_alias(self):
        code = self.get_code()
        def replacer(match):
            return match.group(0).replace(match.group(1), self.get_alias())
        return re.sub(r"workflow\s*(\w+)\s*\{", replacer, code)
    
    def get_code_with_alias_and_id(self):
        code = self.get_code()
        def replacer(match):
            return match.group(0).replace(match.group(1), self.get_alias_and_id())
        from .constant import SUBWORKFLOW_HEADER
        return re.sub(SUBWORKFLOW_HEADER, replacer, code)

    def get_code_to_simplify(self):
        return self.get_code_with_alias_and_id()
    
    def simplify_code(self):
        code = super().simplify_code()

        for o in self.emit:
            code = code.replace(o.get_code(get_OG = True), o.simplify_code(return_tab = False), 1)

        #Renaming the takes in the subworkflow (only body)
        code_up_to_emit, code_after_emit = code, ""
        for match in re.finditer(constant.EMIT_SUBWORKFLOW, code):
            start, _ = match.span(0)
            code_up_to_emit = code[:start]
            code_after_emit = code[start:]
        name = self.get_alias()
        for t in self.take:
            if(len(t.get_gives())!=1):
                raise Exception("This shoudn't happen")
            ch = t.get_gives()[0]
            code_up_to_emit = replace_group1(code_up_to_emit, fr"[^\w]({re.escape(ch.get_code())})[^\w]", f"{ch.get_code()}_{name}")



        #Renaming the emits so that the name of the subworkflow appears in the param name
        #The renaming of the emits outside the subworkflow is done at the emit level (when we simplify the emit)
        for e in self.emit:
            if(len(e.gives)==1):
                ch = e.gives[0]
                if(ch.get_type()=="Channel"):
                    temp = code_after_emit
                    code_after_emit = replace_group1(code_after_emit, fr"[^\w]({re.escape(ch.get_code())})[^\w]", f"{ch.get_code()}_{name}")
                    if(temp==code_after_emit):
                        raise Exception("This shoudn't happen -> code hasn't been replaced")
                elif(ch.get_type()=="Emitted"):
                    #Case it's an emitted -> we don't change anything -> it is already unique (since the ID has been added to the call)
                    None
                else:
                    raise Exception("This shouldn't happen")
            else:
                ch = e.origins[0]
                if(ch.get_type()=="Channel"):
                    temp = code_after_emit
                    code_after_emit = replace_group1(code_after_emit, fr"[^\w]({re.escape(ch.get_code())})[^\w]", f"{ch.get_code()}_{name}")
                    code_up_to_emit+=f"\n{ch.get_code()}_{name} = {ch.get_code()}"
                    if(temp==code_after_emit):
                        raise Exception("This shoudn't happen -> code hasn't been replaced")
                elif(ch.get_type()=="Emitted"):
                    #Case it's an emitted -> we don't change anything -> it is already unique (since the ID has been added to the call)
                    None
                else:
                    raise Exception("This shouldn't happen")
        


        #In the emits replacing the 'ch = ....' by 'ch' and adding 'ch = ....' at the end of the body
        to_replace = []
        for match in re.finditer(r"(\w+) *= *.+", code_after_emit):
            old, new = match.group(0), match.group(1)
            to_replace.append((old, new))
        for r in to_replace:
            old, new = r
            code_up_to_emit+=f"\n{old}"
            temp = code_after_emit
            code_after_emit = code_after_emit.replace(old, new)
            if(temp==code_after_emit):
                raise Exception("This shoudn't happen -> code hasn't been replaced")

        ##Renaming the takes in the emits -> if the takes are given as emits
        #for e in self.emit:
        #    channels_take = []
        #    for t in self.take:
        #        channels_take.append(t.get_gives()[0])
        #    re_write_channel = False
        #    for o in e.origins:
        #        if(o in channels_take):
        #            re_write_channel = True
        #    if(re_write_channel):
        #        ch = e.origins[0]
        #        temp = code_after_emit
        #        code_after_emit = replace_group1(code_after_emit, fr"[^\w]({re.escape(ch.get_code())})[^\w]", f"{ch.get_code()}_{name}")
        #        if(temp==code_after_emit):
        #            raise Exception("This shoudn't happen -> code hasn't been replaced")

        code = code_up_to_emit+'\n'+code_after_emit
        return code


    def set_alias(self, alias):
        self.alias = alias

    def set_printed_name(self, name):
        self.printed_name = name

    def get_alias(self):
        return self.alias

    def get_alias_and_id(self):
        return f"{self.alias}_GG_{id(self)}"
    
    def get_printed_name(self):
        return self.printed_name

    def get_type(self):
        return "Subworkflow"

    def get_name(self):
        return self.name
    
    def get_work(self):
        return self.work.get_code()
    
    #TODO -> when return the code of a subworkflow -> i return evrything (not just the work) -> check if that is correct
    #Method which initiliases the different parts of a workflow (take/main/emit)
    def initialise_parts(self):
        code = self.get_code()
        take_multiple, take_pos= False, (0, 0)
        for match in re.finditer(constant.TAKE, code):
            if(take_multiple):
                raise BioFlowInsightError(f"Multiple 'take:' were found in the subworkflow '{self.get_name()}'.", type = 22, origin=self)
            take_pos = match.span(0)
            take_multiple = True

        main_multiple, main_pos= False, (0, 0)
        for match in re.finditer(constant.MAIN, code):
            if(main_multiple):
                raise BioFlowInsightError(f"Multiple 'main:' were found in the subworkflow '{self.get_name()}'.", type = 22, origin=self)
            main_pos = match.span(0)
            main_multiple = True

        emit_multiple, emit_pos= False, (0, 0)
        for match in re.finditer(constant.EMIT_SUBWORKFLOW, code):
            if(emit_multiple):
                raise BioFlowInsightError(f"Multiple 'emit:' were found in the subworkflow '{self.get_name()}'. ", type = 22, origin=self)
            emit_pos = match.span(0)
            emit_multiple = True

        #Case everything is there
        if(take_pos!=(0, 0) and main_pos!=(0, 0) and emit_pos!=(0, 0)):
            if(take_pos[0]<main_pos[0] and main_pos[0]<emit_pos[0]):
                self.take = Code(code[take_pos[1]:main_pos[0]].strip(), origin = self, initialise=False)
                self.work = Code(code[main_pos[1]:emit_pos[0]].strip(), origin = self, initialise=False)
                self.emit = Code(code[emit_pos[1]:code.rfind('}')].strip(), origin = self, initialise=False)
            elif(take_pos[0]<emit_pos[0] and emit_pos[0]<main_pos[0]):
                self.take = Code(code[take_pos[1]:emit_pos[0]].strip(), origin = self, initialise=False)
                self.emit = Code(code[emit_pos[1]:main_pos[0]].strip(), origin = self, initialise=False)
                self.work = Code(code[main_pos[1]:code.rfind('}')].strip(), origin = self, initialise=False)
            else:
                raise Exception('You need to add a case')
        #Case nothing is there
        if(take_pos==(0, 0) and main_pos==(0, 0) and emit_pos==(0, 0)):
            #raise Exception(f"Subworkflow {code} doesn't have anything defined")
            firt_curly  = code.find("{")
            last_curly = code.rfind('}')
            self.work = Code(code[firt_curly+1:last_curly], origin = self, initialise=False)
        #Case there is an input but no output
        if(take_pos!=(0, 0) and main_pos!=(0, 0) and emit_pos==(0, 0)):
            if(take_pos[0]<main_pos[0]):
                self.take = Code(code[take_pos[1]:main_pos[0]].strip(), origin = self, initialise=False)
                self.work = Code(code[main_pos[1]:code.rfind('}')].strip(), origin = self, initialise=False)
            else:
                raise Exception('You need to add a case')
        #Case there is no input but an output
        if(take_pos==(0, 0) and main_pos!=(0, 0) and emit_pos!=(0, 0)):
            if(main_pos[0]<emit_pos[0]):
                self.work = Code(code[main_pos[1]:emit_pos[0]].strip(), origin = self, initialise=False)
                self.emit = Code(code[emit_pos[1]:code.rfind('}')].strip(), origin = self, initialise=False)
            else:
                raise Exception('You need to add a case')
        #Case there is a main but no input and no output
        if(take_pos==(0, 0) and main_pos!=(0, 0) and emit_pos==(0, 0)):
            self.work = Code(code[main_pos[1]:code.rfind('}')].strip(), origin = self, initialise=False)
        if( main_pos==(0, 0) and (take_pos!=(0, 0) or emit_pos!=(0, 0))):
            if(take_pos!=(0, 0) and emit_pos!=(0, 0)):
                raise Exception("TODO")
            elif(take_pos!=(0, 0) and emit_pos==(0, 0)):
                raise Exception("TODO")
            elif(take_pos==(0, 0) and emit_pos!=(0, 0)):
                self.emit = Code(code[emit_pos[1]:code.rfind('}')].strip(), origin = self, initialise=False)
                firt_curly  = code.find("{")
                self.work = Code(code[firt_curly+1:emit_pos[0]].strip(), origin = self, initialise=False)
            else:
                raise Exception("Not possible!")
    
    def get_channel_from_name_takes(self, name):
        for c in self.takes_channels:
            if(name == c.get_name()):
                return c
        return None

    def initialise_takes(self):
        if(self.take!=[]):
            code = remove_jumps_inbetween_parentheses(self.take.get_code()).split('\n')
            tab = []
            for i in range(len(code)):
                code[i] = code[i].strip()
                if(code[i]!=''):
                    channel = self.get_channel_from_name_takes(code[i])
                    #channel = self.root.get_channel_from_name(code[i])
                    #In the case the channel doesn't exist
                    if(channel==None):
                        from .operation import Operation
                        ope = Operation(f"t: {code[i]}", self)
                        ope.set_as_artificial()
                        from .channel import Channel
                        channel = Channel(code[i], self)
                        ope.add_element_gives(channel)
                        channel.add_source(ope)
                        #ope.initialise_from_subworkflow_take()
                    else:
                        raise BioFlowInsightError(f"The channel '{code[i]}' is already defined somewhere else in the subworkflow ('{self.get_name()}') or in the file.", type=4, origin=self)
                    tab.append(ope)
                    for channel in ope.get_gives():
                        self.takes_channels.append(channel)
            self.take = tab




    def initialise_emit(self):
        from .operation import Operation
        if(self.emit!=[]):
            code = remove_jumps_inbetween_parentheses(self.emit.get_code()).split('\n')
            tab = []
            for i in range(len(code)):
                code[i] = code[i].strip()
                if(code[i]!=""):
                    channels = self.root.get_channels_from_name_same_level(code[i])
                    if(channels==[]):
                        channels = self.root.get_channels_from_name_inside_level(code[i])
                    if(channels==[]):
                        channels = self.root.get_channels_from_name_other_blocks_on_same_level(code[i])
                    if(channels!=[]):
                        ope = Operation(code=f"e: {code[i]}", origin=self)
                        ope.set_as_artificial()
                        for channel in channels:
                            ope.add_element_origins(channel)
                            channel.add_sink(ope)
                        tab.append(ope)
                        
                    else:
                        #raise Exception(f"I don't know how to handle '{code[i]}'")
                        #Case it's an operation 
                        operation = Operation(code[i], self)
                        operation.initialise()
                        operation.change_code(f"e: {code[i]}")
                        operation.set_as_artificial()
                        tab.append(operation)
                        #operation.add_gives(channel)
                        #for gives in operation.get_gives():
                        #    #TODO -> check not add origin too!
                        #    gives.add_sink(operation)
                        #tab.append(operation)
                        ##self.add_operation(operation)
                        ##self.executors.append(operation)
            self.emit = tab
            
            

    

    def get_emit(self):
        return self.emit
    
    def get_nb_emit(self):
        return len(self.emit)

    def get_takes(self):
        return self.take
    
    def get_nb_takes(self):
        return len(self.take)
    
    def get_nb_inputs(self):
        return self.get_nb_takes()
            

    def initialise(self):
        if(not self.initialised):
            self.initialised = True
            self.initialise_parts()
            self.initialise_takes()
            self.modules_defined = self.nextflow_file.get_modules_defined()
            #Check that includes are not defined in the main or subworkflows
            self.check_includes()
            self.root = Root(code=self.get_work(), origin=self, modules_defined=self.modules_defined, subworkflow_inputs = self.takes_channels)
            self.root.initialise()
            self.initialise_emit()
            

    def get_all_executors_in_subworkflow(self):
        dico = {}
        self.root.get_all_executors_in_subworkflow(calls = dico)
        for operation in self.emit:
            dico[operation] = ""
        return list(dico.keys())

    def get_structure(self, dico):
        super().get_structure(dico)

        for ope in self.get_takes():
            #ope.set_operation_type("Branch")
            ope.get_structure(dico)
        
        for ope in self.get_emit():
            #ope.set_operation_type("Branch")
            ope.get_structure(dico)

  